# pylint: disable=no-self-argument, no-else-return, unused-argument, no-self-use, unused-import
import logging
from typing import Dict, FrozenSet, List

import git

from gto.index import ObjectCommits

from .base import BaseLabel, BaseManager, BaseRegistryState
from .config import CONFIG  # need to pass this when you initialize BranchEnvManager
from .constants import Action


def traverse_commit_check(commit: git.Commit, desired: git.Commit):
    if commit == desired:
        return True
    return any(traverse_commit_check(c, desired) for c in commit.parents)


def find_branches(repo: git.Repo, desired: str) -> List[git.Head]:
    return [
        branch
        for branch in repo.heads
        if traverse_commit_check(branch.commit, repo.commit(desired))
    ]


class BranchEnvManager(BaseManager):
    actions: FrozenSet[Action] = frozenset((Action.PROMOTE, Action.DEMOTE))

    def update_state(
        self, state: BaseRegistryState, index: ObjectCommits
    ) -> BaseRegistryState:
        if CONFIG.VERSION_REQUIRED_FOR_ENV:
            # we assume that the model is promoted the same moment it is registered
            for name in state.objects:
                for version in state.objects[name].versions:
                    # TODO: For each branch that has this commit in history
                    # we assume the model was promoted to corresponding env
                    # we should see are there any options to do this differently
                    for branch in find_branches(self.repo, version.commit_hexsha):
                        # figure out env from branch name
                        env = CONFIG.branch_to_env(branch.name)
                        if env is None:
                            continue
                        state.objects[name].labels.append(
                            BaseLabel(
                                object=version.object,
                                version=version.name,
                                name=env,
                                creation_date=version.creation_date,
                                author=version.author,
                                commit_hexsha=version.commit_hexsha,
                                unregistered_date=version.unregistered_date,
                            )
                        )
        else:
            # we assume each commit in a branch is a promotion to branch env
            # if object was indexed
            for name, commits in index.items():
                for hexsha in commits:
                    commit = self.repo.commit(hexsha)
                    version = state.objects[name].find_version(commit_hexsha=hexsha)  # type: ignore
                    version = version.name if version else hexsha  # type: ignore
                    for branch in find_branches(self.repo, hexsha):
                        env = CONFIG.branch_to_env(branch.name)
                        if env is None:
                            continue
                        state.objects[name].labels.append(
                            BaseLabel(
                                object=name,
                                version=version,
                                name=env,
                                creation_date=commit.committed_date,
                                author=commit.author.name,
                                commit_hexsha=commit.hexsha,
                                unregistered_date=None,
                            )
                        )
        return state

    def promote(
        self,
        name,
        label,
        ref,
        message=None,  # arg is ignored
    ):
        if CONFIG.VERSION_REQUIRED_FOR_ENV:
            # to promote, we need to register the version
            # with this setting, the versions should be already registered in BaseRegistry.promote
            # so we should not need to do anything here, except for maybe checking that it's done

            # special case: if you are trying to promote a commit
            # which has a version already registered and you don't want to deprecate it,
            # then you need to create new commit with a model you want to use and register a version there
            return None

        # to promote, we don't need anything
        if self.repo.heads[CONFIG.env_to_branch(label)].commit.hexsha == ref:
            logging.info(
                "HEAD commit in BRANCH is promoted to ENV by default, so you don't need to run 'promote' command"
            )
            return None
        # except maybe creating a commit in the branch (do we want this?):
        # "gto promote $MODEL $PATH staging" -> create commit in branch for "staging"
        raise NotImplementedError(
            "If you want to promote a REF which is not HEAD of the appropriate branch, "
            "you need to create a new commit to that branch "
            "or move HEAD of it to the REF you want to promote"
        )

    def demote(self, name, label, message=None):
        if CONFIG.VERSION_REQUIRED_FOR_ENV:
            # can be done by deprecating a version. Need something like --deprecate_version flag for CLI
            # to acknowledge the actor understands the implication
            raise NotImplementedError("To demote, you need to deprecate a version")

        # can be done by reversing commit. Need something like --reverse_commit flag for CLI?
        # that will generate a commit with the model from the previous commit
        raise NotImplementedError(
            "To demote, you need to reverse a commit, "
            "move HEAD of the branch to the previous commit, "
            "or create a new commit with the model from the previous commit"
        )

    def check_ref(self, ref: str, state: BaseRegistryState) -> Dict[str, BaseLabel]:
        # we assume ref is a commit. If it's a tag then we don't need to return anything
        # this is my assumption that should be discussed
        # it's based on case when CI will be triggered twice - for registration tag and promotion commit
        # VERSION_BASE='tag' VERSION_REQUIRED_FOR_ENV=True ENV_BASE='branch'
        try:
            assert all(r.name != ref for r in self.repo.refs)
            ref = self.repo.commit(ref).hexsha
        except (git.BadName, AssertionError):
            logging.warning("Reference is not a commit hexsha or it doesn't exist")
            return {}
        return {
            name: label
            for name in state.objects
            for label in state.objects[name].labels
            if label.commit_hexsha == ref
        }



import os
from typing import Union

import click
import git
from git import Repo
from pydantic import BaseModel

from gto.base import BaseManager, BaseObject, BaseRegistryState
from gto.config import CONFIG_FILE, RegistryConfig
from gto.exceptions import (
    NoActiveLabel,
    VersionAlreadyRegistered,
    VersionExistsForCommit,
    VersionIsOld,
)
from gto.index import RepoIndexManager


class GitRegistry(BaseModel):
    repo: git.Repo
    version_manager: BaseManager
    env_manager: BaseManager
    config: RegistryConfig

    class Config:
        arbitrary_types_allowed = True

    @classmethod
    def from_repo(cls, repo=Union[str, Repo], config=None):
        if isinstance(repo, str):
            repo = git.Repo(repo)
        if config is None:
            config = RegistryConfig(
                CONFIG_FILE=os.path.join(repo.working_dir, CONFIG_FILE)
            )

        return cls(
            repo=repo,
            version_manager=config.VERSION_MANAGERS_MAPPING[config.VERSION_BASE](
                repo=repo
            ),
            env_manager=config.ENV_MANAGERS_MAPPING[config.ENV_BASE](repo=repo),
            config=config,
        )

    @property
    def index(self):
        return RepoIndexManager(repo=self.repo)

    @property
    def state(self):
        index = self.index.object_centric_representation()
        state = BaseRegistryState(
            objects={
                name: BaseObject(name=name, versions=[], labels=[]) for name in index
            }
        )
        state = self.version_manager.update_state(state, index)
        state = self.env_manager.update_state(state, index)
        state.sort()
        return state

    def register(self, name, ref, version=None, bump=None):
        """Register object version"""
        ref = self.repo.commit(ref).hexsha
        # TODO: add the same check for other actions, to promote and etc
        # also we need to check integrity of the index+state
        self.index.assert_existence(name, ref)
        found_object = self.state.find_object(name)
        # check that this commit don't have a version already
        found_version = found_object.find_version(
            commit_hexsha=ref, skip_unregistered=True
        )
        if found_version is not None:
            raise VersionExistsForCommit(name, found_version.name)
        # if version name is provided, use it
        if version:
            if (
                found_object.find_version(name=version, skip_unregistered=False)
                is not None
            ):
                raise VersionAlreadyRegistered(version)
            if (
                found_object.versions
                and self.config.versions_class(version)
                < found_object.latest_version.name
            ):
                raise VersionIsOld(
                    latest=found_object.latest_version.name, suggested=version
                )
        # if version name wasn't provided but there were some, bump the last one
        elif found_object.versions:
            version = (
                self.config.versions_class(
                    self.state.find_object(name).latest_version.name
                )
                .bump(**({"part": bump} if bump else {}))
                .version
            )
        # if no versions exist, use the minimal version possible
        else:
            version = self.config.versions_class.get_minimal().version
        self.version_manager.register(
            name,
            version,
            ref,
            message=f"Registering object {name} version {version}",
        )
        return self.state.find_object(name).find_version(
            name=version, raise_if_not_found=True
        )

    def unregister(self, name, version):
        return self.version_manager.unregister(name, version)

    def promote(
        self,
        name,
        label,
        promote_version=None,
        promote_ref=None,
        name_version=None,
    ):
        """Assign label to specific object version"""
        self.config.assert_env(label)
        if not (promote_version is None) ^ (promote_ref is None):
            raise ValueError("One and only one of (version, commit) must be specified.")
        if promote_ref:
            promote_ref = self.repo.commit(promote_ref).hexsha
        if promote_ref:
            self.index.assert_existence(name, promote_ref)
        found_object = self.state.find_object(name)
        if promote_version is not None:
            found_version = found_object.find_version(
                name=promote_version, raise_if_not_found=True
            )
            promote_ref = self.find_commit(name, promote_version)
        else:
            found_version = found_object.find_version(commit_hexsha=promote_ref)
            if found_version is None:
                self.register(name, version=name_version, ref=promote_ref)
                click.echo(
                    f"Registered new version '{promote_version}' of '{name}' at commit '{promote_ref}'"
                )
        self.env_manager.promote(
            name,
            label,
            ref=promote_ref,
            message=f"Promoting {name} version {promote_version} to label {label}",
        )
        return {"version": promote_version}

    def demote(self, name, label):
        """De-promote object from given label"""
        # TODO: check if label wasn't demoted already
        if self.state.find_object(name).latest_labels.get(label) is None:
            raise NoActiveLabel(label=label, name=name)
        return self.env_manager.demote(
            name,
            label,
            message=f"Demoting {name} from label {label}",
        )

    def check_ref(self, ref: str):
        "Find out what was registered/promoted in this ref"
        return {
            "version": self.version_manager.check_ref(ref, self.state),
            "env": self.env_manager.check_ref(ref, self.state),
        }

    def find_commit(self, name, version):
        return self.state.find_commit(name, version)

    def which(self, name, label, raise_if_not_found=True):
        """Return label active in specific env"""
        return self.state.which(name, label, raise_if_not_found)

    def latest(self, name: str):
        """Return latest version for object"""
        return self.state.find_object(name).latest_version



from collections.abc import Iterable
from copy import deepcopy
from datetime import datetime


def flatten(obj):
    if isinstance(obj, Iterable) and not isinstance(obj, (str, bytes)):
        return "_".join(obj)
    return obj


def serialize(data_to_serialize):
    data = deepcopy(data_to_serialize)
    if isinstance(data, (int, float, str, bool)):
        return data
    if isinstance(data, datetime):
        return data.isoformat()
    if isinstance(data, list):
        return [serialize(i) for i in data]
    if isinstance(data, dict):
        return {flatten(key): serialize(value) for key, value in data.items()}
    if data is None:
        return data
    raise NotImplementedError(
        f"Serialisation is not implemented for {data_to_serialize} of type {type(data_to_serialize)}"
    )



import json
import logging
import warnings
from functools import wraps

import click
import pandas as pd
from ruamel import yaml

import gto
from gto.constants import LABEL, NAME, REF, VERSION
from gto.utils import serialize

arg_name = click.argument(NAME)
arg_version = click.argument(VERSION)
arg_label = click.argument(LABEL)
arg_ref = click.argument(REF)
option_repo = click.option("-r", "--repo", default=".", help="Repository to use")
option_format = click.option("--format", "-f", default="yaml", help="Output format")


@click.group()
def cli():
    """\b
    Great Tool Ops. Turn your Git Repo into Artifact Registry:
    * Index your artifacts and add enrichments
    * Register artifact versions
    * Promote artifacts to environments
    * Act on new versions and promotions in CI
    """


def _set_log_level(ctx, param, value):  # pylint: disable=unused-argument
    if value:
        logger = logging.getLogger("gto")
        logger.handlers[0].setLevel(logging.DEBUG)
        logger.setLevel(logging.DEBUG)
        from gto.config import CONFIG  # pylint: disable=import-outside-toplevel

        click.echo(CONFIG)


verbose_option = click.option(
    "-v",
    "--verbose",
    callback=_set_log_level,
    expose_value=False,
    is_eager=True,
    is_flag=True,
)


@wraps(cli.command)
def gto_command(*args, **kwargs):
    def decorator(func):
        return cli.command(*args, **kwargs)(verbose_option(func))

    return decorator


@gto_command()
@option_repo
@arg_name
@click.argument("type")
@click.argument("path")
def add(repo: str, name: str, type: str, path: str):
    """Add an object to the Index"""
    gto.api.add(repo, name, type, path)


@cli.command("rm")
@option_repo
@arg_name
def remove(repo: str, name: str):
    """Remove an object from the Index"""
    gto.api.remove(repo, name)


@gto_command()
@option_repo
@arg_name
@arg_ref
@click.option("--version", "-v", default=None, help="Version to promote")
@click.option(
    "--bump", "-b", default=None, help="The exact part to use when bumping a version"
)
def register(repo: str, name: str, ref: str, version: str, bump: str):
    """Register new object version"""
    registered_version = gto.api.register(
        repo=repo, name=name, ref=ref, version=version, bump=bump
    )
    click.echo(
        f"Registered {registered_version.object} version {registered_version.name}"
    )


@gto_command()
@option_repo
@arg_name
@arg_version
def unregister(repo: str, name: str, version: str):
    """Unregister object version"""
    gto.api.unregister(repo, name, version)
    click.echo(f"Unregistered {name} version {version}")


@gto_command()
@option_repo
@arg_name
@arg_label
@click.option(
    "--version",
    default=None,
    help="If you provide --commit, this will be used to name new version",
)
@click.option("--ref", default=None)
def promote(repo: str, name: str, label: str, version: str, ref: str):
    """Assign label to specific object version"""
    if ref is not None:
        name_version = version
        promote_version = None
    else:
        name_version = None
        promote_version = version
    result = gto.api.promote(repo, name, label, promote_version, ref, name_version)
    click.echo(f"Promoted {name} version {result['version']} to label {label}")


@gto_command()
@option_repo
@arg_name
def latest(repo: str, name: str):
    """Return latest version for object"""
    latest_version = gto.api.find_latest_version(repo, name)
    if latest_version:
        click.echo(latest_version.name)
    else:
        click.echo("No versions found")


@gto_command()
@option_repo
@arg_name
@arg_label
def which(repo: str, name: str, label: str):
    """Return version of object with specific label active"""
    version = gto.api.find_active_label(repo, name, label)
    if version:
        click.echo(version.version)
    else:
        click.echo(f"No version of '{name}' with label '{label}' active")


@gto_command()
@option_repo
@arg_name
@arg_label
def demote(repo: str, name: str, label: str):
    """De-promote object from given label"""
    gto.api.demote(repo, name, label)
    click.echo(f"Demoted {name} from label {label}")


@gto_command()
@arg_name
@click.option("--key", default=None, help="Which key to return")
def parse_tag(name: str, key: str):
    parsed = gto.api.parse_tag(name)
    if key:
        parsed = parsed[key]
    click.echo(parsed)


@gto_command()
@option_repo
@click.argument("ref")
@option_format
def check_ref(repo: str, ref: str, format: str):
    """Find out what have been registered/promoted in the provided ref"""
    result = gto.api.check_ref(repo, ref)
    if format == "yaml":
        click.echo(yaml.dump(result, default_style='"'))
    elif format == "json":
        click.echo(json.dumps(serialize(result)))
    else:
        raise NotImplementedError("Unknown format")


@gto_command()
@option_repo
@click.option("--format", "-f", default="dataframe", help="Output format")
def show(repo: str, format: bool):
    """Show current registry state"""
    if format == "dataframe":
        click.echo(gto.api.show(repo, dataframe=True))
    elif format == "json":
        click.echo(json.dumps(gto.api.show(repo, dataframe=False)))
    elif format == "yaml":
        click.echo(
            yaml.dump(gto.api.show(repo, dataframe=False), default_flow_style=False)
        )
    else:
        raise NotImplementedError("Unknown format")


@gto_command()
@click.argument("action")
@option_repo
def audit(action: str, repo: str):
    """Audit registry state"""

    if action in {"reg", "registration", "register", "all"}:
        click.echo("\n=== Registration audit trail ===")
        click.echo(gto.api.audit_registration(repo, dataframe=True))

    if action in {"promote", "promotion", "all"}:
        click.echo("\n=== Promotion audit trail ===")
        click.echo(gto.api.audit_promotion(repo, dataframe=True))


@gto_command()
@option_repo
@option_format
def print_state(repo: str, format: str):
    """Print current registry state"""
    state = serialize(gto.api.get_state(repo).dict())
    if format == "yaml":
        click.echo(yaml.dump(serialize(state), default_flow_style=False))
    elif format == "json":
        click.echo(json.dumps(state))
    else:
        raise NotImplementedError("Unknown format")


@gto_command()
@option_repo
@option_format
def print_index(repo: str, format: str):
    index = gto.api.get_index(repo).object_centric_representation()
    if format == "yaml":
        click.echo(
            yaml.dump(
                dict(index),
                default_flow_style=False,
            )
        )
    elif format == "json":
        click.echo(json.dumps(index))
    else:
        raise NotImplementedError("Unknown format")


if __name__ == "__main__":
    warnings.filterwarnings("ignore")
    pd.set_option("display.max_colwidth", 100)

    cli()



from typing import Union

import pandas as pd
from git import Repo

from gto.index import FileIndexManager, RepoIndexManager
from gto.registry import GitRegistry
from gto.tag import parse_name


def get_index(repo: Union[str, Repo], file=False):
    """Get index state"""
    if file:
        return FileIndexManager(
            path=repo if isinstance(repo, str) else repo.working_dir
        )
    return RepoIndexManager.from_repo(repo)


def get_state(repo: Union[str, Repo]):
    """Show current registry state"""
    return GitRegistry.from_repo(repo).state


def add(repo: Union[str, Repo], name: str, type: str, path: str):
    """Add an object to the Index"""
    return FileIndexManager(path=repo).add(name, type, path)


def remove(repo: Union[str, Repo], name: str):
    """Remove an object from the Index"""
    return FileIndexManager(path=repo).remove(name)


def register(
    repo: Union[str, Repo], name: str, ref: str, version: str = None, bump: str = None
):
    """Register new object version"""
    return GitRegistry.from_repo(repo).register(
        name=name, ref=ref, version=version, bump=bump
    )


def unregister(repo: Union[str, Repo], name: str, version: str):
    """Unregister object version"""
    return GitRegistry.from_repo(repo).unregister(name, version)


def promote(
    repo: Union[str, Repo],
    name: str,
    label: str,
    promote_version: str = None,
    promote_ref: str = None,
    name_version: str = None,
):
    """Assign label to specific object version"""
    return GitRegistry.from_repo(repo).promote(
        name, label, promote_version, promote_ref, name_version
    )


def demote(repo: Union[str, Repo], name: str, label: str):
    """De-promote object from given label"""
    return GitRegistry.from_repo(repo).demote(name, label)


def parse_tag(name: str):
    return parse_name(name)


def find_latest_version(repo: Union[str, Repo], name: str):
    """Return latest version for object"""
    return GitRegistry.from_repo(repo).latest(name)


def find_active_label(repo: Union[str, Repo], name: str, label: str):
    """Return version of object with specific label active"""
    return GitRegistry.from_repo(repo).which(name, label, raise_if_not_found=False)


def check_ref(repo: Union[str, Repo], ref: str):
    """Find out what have been registered/promoted in the provided ref"""
    reg = GitRegistry.from_repo(repo)
    ref = ref.removeprefix("refs/tags/")
    if ref.startswith("refs/heads/"):
        ref = reg.repo.commit(ref).hexsha
    result = reg.check_ref(ref)
    return {
        action: {name: version.dict() for name, version in found.items()}
        for action, found in result.items()
    }


def show(repo: Union[str, Repo], dataframe: bool = False):
    """Show current registry state"""
    reg = GitRegistry.from_repo(repo)
    models_state = {
        o.name: {
            "version": o.latest_version.name if o.latest_version else None,
            "environment": {
                l: o.latest_labels[l].version
                if o.latest_labels[l] is not None
                else None
                for l in o.unique_labels
            },
        }
        for o in reg.state.objects.values()
    }
    if dataframe:
        result = {
            ("", "latest"): {name: d["version"] for name, d in models_state.items()}
        }
        for name, details in models_state.items():
            for env, ver in details["environment"].items():
                result[("environment", env)] = {
                    **result.get(("environment", env), {}),
                    **{name: ver},
                }
        result_df = pd.DataFrame(result)
        result_df.columns = pd.MultiIndex.from_tuples(result_df.columns)
        return result_df
    return models_state


def audit_registration(repo: Union[str, Repo], dataframe: bool = False):
    """Audit registry state"""
    reg = GitRegistry.from_repo(repo)

    audit_trail = [
        {
            "name": o.name,
            "version": v.name,
            "creation_date": v.creation_date,
            "author": v.author,
            "commit_hexsha": v.commit_hexsha,
            "unregistered_date": v.unregistered_date,
        }
        for o in reg.state.objects.values()
        for v in o.versions
    ]
    if not dataframe:
        return audit_trail

    df = pd.DataFrame(audit_trail)
    if len(df):
        df.sort_values("creation_date", ascending=False, inplace=True)
        df.set_index(["creation_date", "name"], inplace=True)
    return df


def audit_promotion(repo: Union[str, Repo], dataframe: bool = False):
    """Audit registry state"""
    reg = GitRegistry.from_repo(repo)
    audit_trail = [
        {
            "name": o.name,
            "label": l.name,
            "version": l.version,
            "creation_date": l.creation_date,
            "author": l.author,
            "commit_hexsha": l.commit_hexsha,
            "unregistered_date": l.unregistered_date,
        }
        for o in reg.state.objects.values()
        for l in o.labels
    ]
    if not dataframe:
        return audit_trail

    df = pd.DataFrame(audit_trail)
    if len(df):
        df.sort_values("creation_date", ascending=False, inplace=True)
        df.set_index(["creation_date", "name"], inplace=True)
    return df



import logging
from typing import Dict, FrozenSet

from git import BadName

from .base import BaseManager, BaseRegistryState, BaseVersion
from .constants import Action
from .index import ObjectCommits


class CommitVersionManager(BaseManager):
    actions: FrozenSet[Action] = frozenset((Action.PROMOTE, Action.DEMOTE))

    def update_state(
        self, state: BaseRegistryState, index: ObjectCommits
    ) -> BaseRegistryState:
        # each commit is a version if object is indexed in that commit
        for name, commits in index.items():
            for hexsha in commits:
                commit = self.repo.commit(hexsha)
                state.objects[name].versions.append(
                    BaseVersion(
                        object=name,
                        name=hexsha,
                        creation_date=commit.committed_datetime,
                        author=commit.author.name,
                        commit_hexsha=hexsha,
                    )
                )
        return state

    def check_ref(self, ref: str, state: BaseRegistryState) -> Dict[str, BaseVersion]:
        try:
            # check this is a commit and it exists
            assert all(r.name != ref for r in self.repo.refs)
            ref = self.repo.commit(ref).hexsha
        except (BadName, AssertionError):
            logging.warning("Reference is not a commit hexsha or it doesn't exist")
            return {}
        return {
            name: version
            for name in state.objects
            for version in state.objects[name].versions
            if version.commit_hexsha == ref
        }



from functools import total_ordering
from typing import Union

import semver

from gto.constants import VersionPart
from gto.exceptions import GTOException, IncomparableVersions, InvalidVersion


class AbstractVersion:
    version: str

    def __init__(self, version) -> None:
        if not self.is_valid(version):
            raise InvalidVersion(version=version, cls=self.__class__)
        self.version = version

    @classmethod
    def is_valid(cls, version) -> bool:
        raise NotImplementedError

    def __eq__(self, other):
        return self.version == other.version

    def __lt__(self, other):
        raise NotImplementedError

    def bump(self, part: Union[VersionPart, str]) -> "AbstractVersion":
        raise NotImplementedError

    @classmethod
    def get_minimal(cls):
        raise NotImplementedError


@total_ordering
class NumberedVersion(AbstractVersion):
    @classmethod
    def is_valid(cls, version):
        if not isinstance(version, str):
            return False
        return version.startswith("v") and version[1:].isdigit()

    def to_number(self):
        return int(self.version[1:])

    def __eq__(self, other):
        if isinstance(other, str):
            other = self.__class__(other)
        if not isinstance(other, self.__class__):
            raise IncomparableVersions(self, other)
        return self.version == other.version

    def __lt__(self, other):
        if isinstance(other, str):
            other = self.__class__(other)
        if not isinstance(other, self.__class__):
            raise IncomparableVersions(self, other)
        return self.to_number() < other.to_number()

    def bump(self, part: Union[VersionPart, str] = VersionPart.MAJOR):
        part = VersionPart(part)
        if part != VersionPart.MAJOR:
            raise GTOException(f"With {self.__class__} you can only bump MAJOR part")
        return self.__class__(f"v{self.to_number() + 1}")

    @classmethod
    def get_minimal(cls):
        return cls("v1")


@total_ordering
class SemVer(AbstractVersion):
    """
    A subclass of Version which allows a "v" prefix
    """

    @classmethod
    def is_valid(cls, version):
        try:
            cls.parse(version)
            return True
        except ValueError:
            return False

    @classmethod
    def parse(cls, version: str) -> "SemVer":
        """
        Parse version string to a Version instance.

        :param version: version string with "v" or "V" prefix
        :raises ValueError: when version does not start with "v" or "V"
        :return: a new instance
        """
        if not isinstance(version, str):
            raise ValueError("Version should be of type str")
        if version[0] not in ("v"):
            raise ValueError(
                f"{version}: not a valid semantic version tag. Must start with 'v'"
            )
        return semver.VersionInfo.parse(version[1:])

    def __eq__(self, other):
        if isinstance(other, str):
            other = self.__class__(other)
        if not isinstance(other, self.__class__):
            raise IncomparableVersions(self, other)
        return self.version == other.version

    def __lt__(self, other):
        if isinstance(other, str):
            other = self.__class__(other)
        if not isinstance(other, self.__class__):
            raise IncomparableVersions(self, other)
        return self.parse(self.version) < self.parse(other.version)

    def bump(self, part: Union[VersionPart, str] = VersionPart.PATCH):
        part = VersionPart(part)
        next_version = getattr(self.parse(self.version), f"bump_{part.value}")()
        return self.__class__(f"v{next_version}")

    @classmethod
    def get_minimal(cls):
        return cls("v0.0.1")



import os
from abc import ABC, abstractmethod
from collections import defaultdict
from functools import wraps
from pathlib import Path
from typing import IO, Dict, Generator, List, Optional, Union

import git
from pydantic import BaseModel, parse_obj_as
from ruamel.yaml import safe_dump, safe_load

from .config import CONFIG
from .exceptions import GTOException, ObjectNotFound


class Artifact(BaseModel):
    name: str
    path: str
    type: str


State = Dict[str, Artifact]


def not_frozen(func):
    @wraps(func)
    def inner(self: "Index", *args, **kwargs):
        if self.frozen:
            raise ValueError(f"Cannot {func.__name__}: {self.__class__} is frozen")
        return func(self, *args, **kwargs)

    return inner


class Index(BaseModel):
    state: State = {}  # TODO should not be populated until load() is called
    frozen: bool = False

    def __contains__(self, item):
        return item in self.state

    @classmethod
    def read(cls, path_or_file: Union[str, IO], frozen: bool = False):
        index = cls(frozen=frozen)
        index.state = index.read_state(path_or_file)
        return index

    @staticmethod
    def read_state(path_or_file: Union[str, IO]):
        if isinstance(path_or_file, str):
            with open(path_or_file, "r", encoding="utf8") as file:
                return parse_obj_as(State, safe_load(file))
        return parse_obj_as(State, safe_load(path_or_file))

    def write_state(self, path_or_file: Union[str, IO]):
        if isinstance(path_or_file, str):
            with open(path_or_file, "w", encoding="utf8") as file:
                file.write(safe_dump(self.dict()["state"], default_flow_style=False))

    @not_frozen
    def add(self, name, type, path):
        if name in self:
            raise GTOException(f"Artifact {name} already exists")
        self.state[name] = Artifact(name=name, type=type, path=path)

    @not_frozen
    def remove(self, name):
        if name not in self:
            raise GTOException(f"Artifact {name} does not exist")
        del self.state[name]


class BaseIndexManager(BaseModel, ABC):
    current: Optional[Index]

    @abstractmethod
    def get_index(self) -> Index:
        raise NotImplementedError

    @abstractmethod
    def update(self):
        raise NotImplementedError

    @abstractmethod
    def get_history(self) -> Dict[str, Index]:
        raise NotImplementedError

    def add(self, name, type, path):
        index = self.get_index()
        index.add(name, type, path)
        self.update()

    def remove(self, name):
        index = self.get_index()
        index.remove(name)
        self.update()


class FileIndexManager(BaseIndexManager):
    path: str = ""

    def index_path(self):
        return str(Path(self.path) / CONFIG.INDEX)

    def get_index(self) -> Index:
        if os.path.exists(self.index_path()):
            self.current = Index.read(self.index_path())
        if not self.current:
            self.current = Index()
        return self.current

    def update(self):
        if self.current is not None:
            self.current.write_state(self.index_path())

    def get_history(self) -> Dict[str, Index]:
        raise NotImplementedError("Not a git repo: history is not available")


ObjectCommits = Dict[str, List[str]]


class RepoIndexManager(FileIndexManager):
    repo: git.Repo

    @classmethod
    def from_repo(cls, repo: Union[str, git.Repo]):
        if isinstance(repo, str):
            repo = git.Repo(repo)
        return cls(repo=repo)

    def index_path(self):
        # TODO: config should be loaded from repo too
        return os.path.join(os.path.dirname(self.repo.git_dir), CONFIG.INDEX)

    class Config:
        arbitrary_types_allowed = True

    def get_commit_index(self, ref: str) -> Index:
        return Index.read(
            (self.repo.commit(ref).tree / CONFIG.INDEX).data_stream, frozen=True
        )

    def get_history(self) -> Dict[str, Index]:
        commits = {
            commit
            for branch in self.repo.heads
            for commit in traverse_commit(branch.commit)
        }
        return {
            commit.hexsha: self.get_commit_index(commit.hexsha)
            for commit in commits
            if CONFIG.INDEX in commit.tree
        }

    def object_centric_representation(self) -> ObjectCommits:
        representation = defaultdict(list)
        for commit, index in self.get_history().items():
            for obj in index.state:
                representation[obj].append(commit)
        return representation

    def check_existence(self, name, commit):
        return name in self.get_commit_index(commit)

    def assert_existence(self, name, commit):
        if not self.check_existence(name, commit):
            raise ObjectNotFound(name)


def traverse_commit(commit: git.Commit) -> Generator[git.Commit, None, None]:
    yield commit
    for parent in commit.parents:
        yield from traverse_commit(parent)



class GTOException(Exception):
    """Base class for all prototype exceptions."""

    def __init__(self, msg, *args):
        assert msg
        self.msg = msg
        super().__init__(msg, *args)


class ObjectNotFound(GTOException):
    _message = "Requested '{name}' wasn't found in registry"

    def __init__(self, name) -> None:
        self.message = self._message.format(name=name)
        super().__init__(self.message)


class VersionRequired(GTOException):
    _message = "No versions found for '{name}'"

    def __init__(self, name):
        self.message = self._message.format(name=name)
        super().__init__(self.message)


class VersionAlreadyRegistered(GTOException):
    _message = (
        "Version '{version}' already was registered.\n"
        "Even if it was unregistered, you must use another name to avoid confusion."
    )

    def __init__(self, version) -> None:
        self.message = self._message.format(version=version)
        super().__init__(self.message)


class VersionExistsForCommit(GTOException):
    _message = "The model {model} was already registered in this commit with version '{version}'."

    def __init__(self, model, version) -> None:
        self.message = self._message.format(model=model, version=version)
        super().__init__(self.message)


class VersionIsOld(GTOException):
    _message = "Version '{suggested}' is younger than the latest {latest}"

    def __init__(self, latest, suggested) -> None:
        self.message = self._message.format(latest=latest, suggested=suggested)
        super().__init__(self.message)


class UnknownEnvironment(GTOException):
    _message = "Environment '{env}' is not present in your config file. Allowed envs are: {envs}."

    def __init__(self, env) -> None:
        # to avoid circular import
        from .config import CONFIG  # pylint: disable=import-outside-toplevel

        self.message = self._message.format(env=env, envs=CONFIG.ENV_WHITELIST)
        super().__init__(self.message)


class NoActiveLabel(GTOException):
    _message = "No active label '{label}' was found for '{name}'"

    def __init__(self, label, name) -> None:
        self.message = self._message.format(label=label, name=name)
        super().__init__(self.message)


class RefNotFound(GTOException):
    _message = "Ref '{ref}' was not found in the repository history"

    def __init__(self, ref) -> None:
        self.message = self._message.format(ref=ref)
        super().__init__(self.message)


class InvalidVersion(GTOException):
    _message = "Supplied version {version} doesn't look like {cls} version"

    def __init__(self, version, cls) -> None:
        self.message = self._message.format(version=version, cls=cls)
        super().__init__(self.message)


class IncomparableVersions(GTOException):
    _message = "You can compare only versions of the same system, but not {} and {}"

    def __init__(self, this, that) -> None:
        self.message = self._message.format(this, that)
        super().__init__(self.message)


class UnknownAction(GTOException):
    message = "Unknown action '{action}' was requested."

    def __init__(self, action) -> None:
        self.message = self.message.format(action=action)
        super().__init__(self.message)


class MissingArg(GTOException):
    message = "'{arg}' is required."

    def __init__(self, arg) -> None:
        self.message = self.message.format(arg=arg)
        super().__init__(self.message)



import gto.log  # noqa
from gto import api
from gto.config import CONFIG
from gto.index import RepoIndexManager
from gto.registry import GitRegistry

__all__ = ["api", "CONFIG", "RepoIndexManager", "GitRegistry"]



from enum import Enum

REF = "ref"
COMMIT = "commit"
TAG = "tag"
BRANCH = "branch"
FILE = "file"

ACTION = "action"
NAME = "name"
VERSION = "version"
LABEL = "label"
NUMBER = "number"


class Action(Enum):
    REGISTER = "register"
    UNREGISTER = "unregister"
    PROMOTE = "promote"
    DEMOTE = "demote"


class VersionPart(Enum):
    MAJOR = "major"
    MINOR = "minor"
    PATCH = "patch"



import logging
from datetime import datetime
from typing import FrozenSet, Iterable, Optional, Union

import git
from pydantic import BaseModel

from .base import BaseLabel, BaseManager, BaseObject, BaseRegistryState, BaseVersion
from .constants import ACTION, LABEL, NAME, NUMBER, VERSION, Action
from .exceptions import MissingArg, RefNotFound, UnknownAction
from .index import ObjectCommits

ActionSign = {
    Action.REGISTER: "@",
    Action.UNREGISTER: "@!",
    Action.PROMOTE: "#",
    Action.DEMOTE: "#!",
}


def name_tag(
    action: Action,
    name: str,
    version: Optional[str] = None,
    label: Optional[str] = None,
    repo: Optional[git.Repo] = None,
):
    if action in (Action.REGISTER, Action.UNREGISTER):
        return f"{name}{ActionSign[action]}{version}"

    if action in (Action.PROMOTE, Action.DEMOTE):
        if repo is None:
            raise MissingArg(arg="repo")
        numbers = []
        for tag in repo.tags:
            parsed = parse_name(tag.name)
            if parsed[ACTION] in (Action.PROMOTE, Action.DEMOTE):
                numbers.append(parsed[NUMBER])
        new_number = max(numbers) + 1 if numbers else 1
        return f"{name}{ActionSign[action]}{label}-{new_number}"
    raise UnknownAction(action=action)


def parse_name(name: str, raise_on_fail: bool = True):

    # order does matter if you take into account ActionSign values
    for action in (Action.UNREGISTER, Action.REGISTER):
        if ActionSign[action] in name:
            name, version = name.split(ActionSign[action])
            return {
                ACTION: action,
                NAME: name,
                VERSION: version,
            }

    # order does matter if you take into account ActionSign values
    for action in (Action.DEMOTE, Action.PROMOTE):
        if ActionSign[action] in name:
            name, label = name.split(ActionSign[action])
            label, number = label.split("-")
            return {
                ACTION: action,
                NAME: name,
                LABEL: label,
                NUMBER: int(number),
            }
    if raise_on_fail:
        raise ValueError(f"Unknown tag name: {name}")
    return {}


class ObjectTag(BaseModel):
    action: Action
    name: str
    version: Optional[str]
    label: Optional[str]
    creation_date: datetime
    tag: git.Tag

    class Config:
        arbitrary_types_allowed = True


def parse_tag(tag: git.Tag):
    return ObjectTag(
        tag=tag,
        creation_date=datetime.fromtimestamp(tag.tag.tagged_date),
        **parse_name(tag.name),
    )


def find(
    action: Union[Action, FrozenSet[Action]] = None,
    name: Optional[str] = None,
    version: Optional[str] = None,
    label: Optional[str] = None,
    repo: Optional[git.Repo] = None,
    sort: str = "by_time",
    tags: Optional[Iterable[git.Tag]] = None,
):
    if isinstance(action, Action):
        action = frozenset([action])
    if tags is None:
        if repo is None:
            raise MissingArg(arg="repo")
        tags = [t for t in repo.tags if parse_name(t.name, raise_on_fail=False)]
    if action:
        tags = [t for t in tags if parse_name(t.name)[ACTION] in action]
    if name:
        tags = [t for t in tags if parse_name(t.name).get(NAME) == name]
    if version:
        tags = [t for t in tags if parse_name(t.name).get(VERSION) == version]
    if label:
        tags = [t for t in tags if parse_name(t.name).get(LABEL) == label]
    if sort == "by_time":
        tags = sorted(tags, key=lambda t: t.tag.tagged_date)
    else:
        raise NotImplementedError(f"Unknown sort: {sort}")
    return tags


def create_tag(repo, name, ref, message):
    if all(c.hexsha != ref for c in repo.iter_commits()):
        raise RefNotFound(ref=ref)

    repo.create_tag(
        name,
        ref=ref,
        message=message,
    )


def version_from_tag(tag: git.Tag) -> BaseVersion:
    mtag = parse_tag(tag)
    return BaseVersion(
        object=mtag.name,
        name=mtag.version,
        creation_date=mtag.creation_date,
        author=tag.tag.tagger.name,
        commit_hexsha=tag.commit.hexsha,
    )


def label_from_tag(tag: git.Tag, obj: BaseObject) -> BaseLabel:
    mtag = parse_tag(tag)
    return BaseLabel(
        object=mtag.name,
        version=obj.find_version(
            commit_hexsha=tag.commit.hexsha, raise_if_not_found=True
        ).name,  # type: ignore
        name=mtag.label,
        creation_date=mtag.creation_date,
        author=tag.tag.tagger.name,
        commit_hexsha=tag.commit.hexsha,
    )


def index_tag(obj: BaseObject, tag: git.Tag) -> BaseObject:
    mtag = parse_tag(tag)
    if mtag.action == Action.REGISTER:
        obj.versions.append(version_from_tag(tag))
    if mtag.action == Action.UNREGISTER:
        obj.find_version(mtag.version).unregistered_date = mtag.creation_date  # type: ignore
    if mtag.action == Action.PROMOTE:
        obj.labels.append(label_from_tag(tag, obj))
    if mtag.action == Action.DEMOTE:
        if mtag.label not in obj.latest_labels:
            raise ValueError(f"Active label '{mtag.label}' not found")
        obj.latest_labels[mtag.label].unregistered_date = mtag.creation_date  # type: ignore
    return obj


class TagManager(BaseManager):  # pylint: disable=abstract-method
    def update_state(
        self, state: BaseRegistryState, index: ObjectCommits
    ) -> BaseRegistryState:
        # tags are sorted and then indexed by timestamp
        # this is important to check that history is not broken
        tags = [parse_tag(t) for t in find(repo=self.repo, action=self.actions)]
        for tag in tags:
            if tag.name not in state.objects:
                state.objects[tag.name] = BaseObject(
                    name=tag.name, versions=[], labels=[]
                )
            state.objects[tag.name] = index_tag(state.objects[tag.name], tag.tag)
        return state


class TagVersionManager(TagManager):
    actions: FrozenSet[Action] = frozenset((Action.REGISTER, Action.UNREGISTER))

    def register(self, name, version, ref, message):
        create_tag(
            self.repo,
            name_tag(Action.REGISTER, name, version=version, repo=self.repo),
            ref=ref,
            message=message,
        )

    def unregister(self, name, version):
        """Unregister object version"""
        # TODO: search in self, move to base
        tags = find(
            action=Action.REGISTER,
            name=name,
            version=version,
            repo=self.repo,
        )
        if len(tags) != 1:
            raise ValueError(f"Found {len(tags)} git tags for {name} version {version}")
        create_tag(
            self.repo,
            name_tag(Action.UNREGISTER, name, version=version, repo=self.repo),
            ref=tags[0].commit.hexsha,
            message=f"Unregistering {name} version {version}",
        )

    def check_ref(self, ref: str, state: BaseRegistryState):
        try:
            _ = self.repo.tags[ref]
            obj_name = parse_name(ref)[NAME]
            version_name = parse_name(ref)[VERSION]
        except (KeyError, ValueError, IndexError):
            logging.warning(
                "Provided ref doesn't exist or it is not a tag that registers a version"
            )
            return {}
        return {
            name: version
            for name in state.objects
            for version in state.objects[name].versions
            if name == obj_name and version.name == version_name
        }


class TagEnvManager(TagManager):
    actions: FrozenSet[Action] = frozenset((Action.PROMOTE, Action.DEMOTE))

    def promote(self, name, label, ref, message):
        create_tag(
            self.repo,
            name_tag(Action.PROMOTE, name, label=label, repo=self.repo),
            ref=ref,
            message=message,
        )

    def demote(self, name, label, message):
        # TODO: search in self, move to base
        promoted_tag = find(
            action=Action.PROMOTE,
            name=name,
            label=label,
            repo=self.repo,
        )[-1]
        create_tag(
            self.repo,
            name_tag(Action.DEMOTE, name, label=label, repo=self.repo),
            ref=promoted_tag.commit.hexsha,
            message=message,
        )

    def check_ref(self, ref: str, state: BaseRegistryState):
        try:
            tag = self.repo.tags[ref]
            _ = parse_name(ref)[LABEL]
            obj_name = parse_name(ref)[NAME]
        except (KeyError, ValueError, IndexError):
            logging.warning(
                "Provided ref doesn't exist or it is not a tag that promotes to an environment"
            )
            return {}
        return {
            name: label
            for name in state.objects
            for label in state.objects[name].labels
            if name == obj_name
            and label.commit_hexsha == tag.commit.hexsha
            and label.creation_date == datetime.fromtimestamp(tag.tag.tagged_date)
        }



from pathlib import Path

from setuptools import find_packages, setup
from setuptools.command.build_py import build_py

install_requires = [
    "gitpython",
    "click",
    "pandas",
    "numpy",
    "pydantic",
    "pyyaml",
    "ruamel.yaml",
    "semver==3.0.0-dev.3",
]


tests = [
    "pytest",
    "pytest-cov",
    "pytest-lazy-fixture==0.6.3",
    "pytest-mock",
    "pylint",
    # we use this to suppress pytest-related false positives in our tests.
    "pylint-pytest",
    # we use this to suppress some messages in tests, eg: foo/bar naming,
    # and, protected method calls in our tests
    "pylint-plugin-utils",
]


setup_args = dict(  # noqa: C408
    name="great-tool-ops",
    version="0.0.5",
    description="Version and deploy your models following GitOps principles",
    long_description=(Path(__file__).parent / "README.md").read_text(encoding="utf8"),
    author="Alexander Guschin",
    author_email="aguschin@iterative.ai",
    download_url="https://github.com/iterative/gto",
    license="Apache License 2.0",
    install_requires=install_requires,
    extras_require={"tests": tests},
    keywords="git gto mlops object registry developer-tools collaboration",
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    packages=find_packages(exclude=["tests"]),
    include_package_data=True,
    entry_points={
        "console_scripts": ["gto = gto.cli:cli"],
    },
    cmdclass={"build_py": build_py},
    zip_safe=False,
)

if __name__ == "__main__":
    setup(**setup_args)



# pylint: disable=no-self-use, no-self-argument, inconsistent-return-statements, invalid-name, import-outside-toplevel
from pathlib import Path
from typing import Any, Dict, List, Optional

import yaml
from pydantic import BaseSettings, validator
from pydantic.env_settings import InitSettingsSource

from gto.versions import AbstractVersion

from .constants import BRANCH, COMMIT, TAG
from .exceptions import UnknownEnvironment

CONFIG_FILE = "gto.yaml"


def _set_location_init_source(init_source: InitSettingsSource):
    def inner(settings: "RegistryConfig"):
        if "CONFIG_FILE" in init_source.init_kwargs:
            settings.__dict__["CONFIG_FILE"] = init_source.init_kwargs["CONFIG_FILE"]
        return {}

    return inner


def config_settings_source(settings: "RegistryConfig") -> Dict[str, Any]:
    """
    A simple settings source that loads variables from a yaml file in GTO DIR
    """

    encoding = settings.__config__.env_file_encoding
    config_file = getattr(settings, "CONFIG_FILE", CONFIG_FILE)
    if not isinstance(config_file, Path):
        config_file = Path(config_file)
    if not config_file.exists():
        return {}
    conf = yaml.safe_load(config_file.read_text(encoding=encoding))

    return {k.upper(): v for k, v in conf.items()} if conf else {}


class RegistryConfig(BaseSettings):
    INDEX: str = "artifacts.yaml"
    VERSION_BASE: str = TAG
    VERSION_CONVENTION: str = "NumberedVersion"
    VERSION_REQUIRED_FOR_ENV: bool = True
    ENV_BASE: str = TAG
    ENV_WHITELIST: List[str] = []
    ENV_BRANCH_MAPPING: Dict[str, str] = {}
    LOG_LEVEL: str = "INFO"
    DEBUG: bool = False
    CONFIG_FILE: Optional[str] = CONFIG_FILE

    @property
    def VERSION_SYSTEM_MAPPING(self):
        from .versions import NumberedVersion, SemVer

        return {"numbers": NumberedVersion, "semver": SemVer}

    @property
    def VERSION_MANAGERS_MAPPING(self):
        from .commit import CommitVersionManager
        from .tag import TagVersionManager

        return {COMMIT: CommitVersionManager, TAG: TagVersionManager}

    @property
    def ENV_MANAGERS_MAPPING(self):
        from .branch import BranchEnvManager
        from .tag import TagEnvManager

        return {TAG: TagEnvManager, BRANCH: BranchEnvManager}

    def assert_env(self, name):
        if not self.check_env(name):
            raise UnknownEnvironment(name)

    def check_env(self, name):
        if self.ENV_BASE == TAG:
            return name in self.ENV_WHITELIST or not self.ENV_WHITELIST
        if self.ENV_BASE == BRANCH:
            return name in self.ENV_BRANCH_MAPPING or not self.ENV_BRANCH_MAPPING

    def branch_to_env(self, branch_name):
        if self.ENV_BRANCH_MAPPING:
            return self.ENV_BRANCH_MAPPING[branch_name]
        return branch_name

    def env_to_branch(self, env_name):
        if self.ENV_BRANCH_MAPPING:
            return {value: key for key, value in self.ENV_BRANCH_MAPPING.items()}[
                env_name
            ]

    class Config:
        env_prefix = "gto_"
        env_file_encoding = "utf-8"

        @classmethod
        def customise_sources(
            cls,
            init_settings,
            env_settings,
            file_secret_settings,
        ):
            return (
                _set_location_init_source(init_settings),
                init_settings,
                env_settings,
                config_settings_source,
                file_secret_settings,
            )

    # @validator("VERSION_REQUIRED_FOR_ENV", always=True)
    # def validate_version_required_for_env(cls, value):
    #     if not value:
    #         raise NotImplementedError
    #     return value

    # TODO: now fails, make this work
    # @validator("ENV_BASE", always=True, pre=True)
    # def validate_env_base(cls, value):
    #     if value not in cls.ENV_MANAGERS_MAPPING:
    #         raise ValueError(f"ENV_BASE must be one of: {cls.ENV_MANAGERS_MAPPING.keys()}")
    #     return value

    @validator("ENV_WHITELIST", always=True)
    def validate_env_whitelist(cls, value, values):
        if values["ENV_BASE"] == BRANCH:
            # logging.warning("ENV_WHITELIST is ignored when ENV_BASE is BRANCH")
            pass
        return value

    @validator("ENV_BRANCH_MAPPING", always=True)
    def validate_env_branch_mapping(
        cls, value: Dict[str, str], values
    ) -> Dict[str, str]:
        if values["ENV_BASE"] != BRANCH:
            # logging.warning("ENV_BRANCH_MAPPING is ignored when ENV_BASE is not BRANCH")
            return value
        if not isinstance(value, dict):
            raise ValueError(
                f"ENV_BRANCH_MAPPING must be a dict, got {type(value)}",
                "ENV_BRANCH_MAPPING",
            )
        if not all(isinstance(k, str) and isinstance(v, str) for k, v in value.items()):
            raise ValueError(
                "ENV_BRANCH_MAPPING must be a dict of str:str", "ENV_BRANCH_MAPPING"
            )
        return value

    @property
    def versions_class(self) -> AbstractVersion:
        if self.VERSION_CONVENTION not in self.VERSION_SYSTEM_MAPPING:
            raise ValueError(f"Unknown versioning system {self.VERSION_CONVENTION}")
        return self.VERSION_SYSTEM_MAPPING[self.VERSION_CONVENTION]


CONFIG = RegistryConfig()



from datetime import datetime
from typing import Dict, FrozenSet, List, Optional

import git
from pydantic import BaseModel

from gto.constants import Action
from gto.index import ObjectCommits

from .exceptions import GTOException, ObjectNotFound


class BaseLabel(BaseModel):  # pylint: disable=too-many-instance-attributes
    # category: str
    object: str
    version: str
    name: str
    creation_date: datetime
    author: str
    commit_hexsha: str
    unregistered_date: Optional[datetime] = None

    def __repr__(self) -> str:
        return f"Label('{self.object}', '{self.version}', '{self.name}')"

    @property
    def is_registered(self):
        return self.unregistered_date is None


class BaseVersion(BaseModel):
    # category: str
    object: str
    name: str
    creation_date: datetime
    author: str
    commit_hexsha: str
    unregistered_date: Optional[datetime] = None

    def __repr__(self) -> str:
        return f"Version('{self.object}', '{self.name}')"

    @property
    def is_registered(self):
        return self.unregistered_date is None


class BaseObject(BaseModel):
    name: str
    versions: List[BaseVersion]
    labels: List[BaseLabel]

    @property
    def unique_labels(self):
        return {l.name for l in self.labels}

    def __repr__(self) -> str:
        versions = ", ".join(f"'{v.name}'" for v in self.versions)
        labels = ", ".join(f"'{l}'" for l in self.unique_labels)
        return f"Object(versions=[{versions}], labels=[{labels}])"

    @property
    def latest_version(self) -> Optional[BaseVersion]:
        if self.versions:
            return sorted(
                (v for v in self.versions if v.is_registered),
                key=lambda x: x.creation_date,
            )[-1]
        return None

    @property
    def latest_labels(self) -> Dict[str, BaseLabel]:
        labels: Dict[str, BaseLabel] = {}
        for label in self.labels:
            # TODO: check that version exists and wasn't demoted???
            # probably this check should be done during State construction
            # as the rules to know it are all there
            if not label.is_registered:
                continue
            if (
                label.name not in labels
                or labels[label.name].creation_date < label.creation_date
            ):
                labels[label.name] = label
        return labels

    def find_version(
        self,
        name: str = None,
        commit_hexsha: str = None,
        raise_if_not_found=False,
        skip_unregistered=True,
    ) -> Optional[BaseVersion]:
        versions = [
            v
            for v in self.versions
            if (v.name == name if name else True)
            and (v.commit_hexsha == commit_hexsha if commit_hexsha else True)
            and (v.unregistered_date is None if skip_unregistered else True)
        ]
        if raise_if_not_found:
            if len(versions) != 1:
                raise GTOException(
                    f"{len(versions)} versions of object {self.name} found"
                    + ", skipping unregistered"
                    if skip_unregistered
                    else ""
                )
            return versions[0]

        if len(versions) > 1:
            raise GTOException(
                f"{len(versions)} versions of object {self.name} found"
                + ", skipping unregistered"
                if skip_unregistered
                else ""
            )
        return versions[0] if versions else None


class BaseRegistryState(BaseModel):
    objects: Dict[str, BaseObject]

    class Config:
        arbitrary_types_allowed = True

    def find_object(self, name):
        obj = self.objects.get(name)
        if not obj:
            raise ObjectNotFound(name)
        return obj

    @property
    def unique_labels(self):
        return sorted({l for o in self.objects.values() for l in o.unique_labels})

    def find_commit(self, name, version):
        return (
            self.find_object(name)
            .find_version(
                name=version, raise_if_not_found=True, skip_unregistered=False
            )
            .commit_hexsha
        )

    def which(self, name, label, raise_if_not_found=True):
        """Return label active in specific env"""
        latest_labels = self.find_object(name).latest_labels
        if label in latest_labels:
            return latest_labels[label]
        if raise_if_not_found:
            raise ValueError(f"Label {label} not found for {name}")
        return None

    def sort(self):
        for name in self.objects:
            self.objects[name].versions.sort(key=lambda x: (x.creation_date, x.name))
            self.objects[name].labels.sort(
                key=lambda x: (x.creation_date, x.version, x.name)
            )


class BaseManager(BaseModel):
    repo: git.Repo
    actions: FrozenSet[Action]

    class Config:
        arbitrary_types_allowed = True

    def update_state(
        self, state: BaseRegistryState, index: ObjectCommits
    ) -> BaseRegistryState:  # pylint: disable=no-self-use
        raise NotImplementedError

    # better to uncomment this for typing, but it breaks the pylint

    # def register(self, name, version, ref, message):
    #     raise NotImplementedError

    # def unregister(self, name, version):
    #     raise NotImplementedError

    # def promote(self, name, label, ref, message):
    #     raise NotImplementedError

    # def demote(self, name, label, message):
    #     raise NotImplementedError

    def check_ref(self, ref: str, state: BaseRegistryState):
        raise NotImplementedError
